@@ -4,29 +4,31 @@ use bstr::BStr;
 use chrono::SubsecRound;
 use flate2::write::{DeflateEncoder, ZlibEncoder};
 use flate2::Compression;
+use log::{debug, warn};
 use rand::{CryptoRng, Rng};
-use smallvec::SmallVec;
 
 use crate::armor;
 use crate::composed::message::decrypt::*;
 use crate::composed::shared::Deserializable;
 use crate::composed::signed_key::SignedSecretKey;
 use crate::composed::StandaloneSignature;
+use crate::crypto::aead::AeadAlgorithm;
 use crate::crypto::hash::HashAlgorithm;
 use crate::crypto::sym::SymmetricKeyAlgorithm;
 use crate::errors::{Error, Result};
 use crate::packet::{
     write_packet, CompressedData, LiteralData, OnePassSignature, Packet,
-    PublicKeyEncryptedSessionKey, Signature, SignatureConfig, SignatureType, Subpacket,
-    SubpacketData, SymEncryptedData, SymEncryptedProtectedData, SymKeyEncryptedSessionKey,
+    PublicKeyEncryptedSessionKey, Signature, SignatureConfig, SignatureType,
+    SignatureVersionSpecific, Subpacket, SubpacketData, SymEncryptedData,
+    SymEncryptedProtectedData, SymKeyEncryptedSessionKey,
 };
 use crate::ser::Serialize;
 use crate::types::{
-    CompressionAlgorithm, KeyId, KeyTrait, KeyVersion, PublicKeyTrait, SecretKeyTrait, StringToKey,
-    Tag,
+    CompressionAlgorithm, EskType, Fingerprint, KeyId, KeyVersion, PkeskVersion, PublicKeyTrait,
+    SecretKeyTrait, StringToKey, Tag,
 };
 
-/// An [OpenPGP message](https://tools.ietf.org/html/rfc4880.html#section-11.3)
+/// An [OpenPGP message](https://www.rfc-editor.org/rfc/rfc9580.html#name-openpgp-messages)
 #[derive(Clone, Debug, PartialEq, Eq)]
 pub enum Message {
     Literal(LiteralData),
@@ -167,12 +169,31 @@ impl Edata {
         }
     }
 
+    /// Transform decrypted data into a message.
+    /// Bails if the packets contain no message or multiple messages.
+    fn process_decrypted(packet_data: &[u8]) -> Result<Message> {
+        let mut messages = Message::from_bytes_many(packet_data);
+        // First message is the one we want to return
+        let Some(message) = messages.next() else {
+            bail!("no valid message found");
+        };
+        let message = message?;
+
+        // The only other message allowed is a padding packet, which will be skipped
+        // by the parser, so check that we have only a single message.
+        if let Some(msg) = messages.next() {
+            bail!("unexpected message: {:?}", msg);
+        }
+
+        Ok(message)
+    }
+
     pub fn decrypt(&self, key: PlainSessionKey) -> Result<Message> {
         let protected = self.tag() == Tag::SymEncryptedProtectedData;
         debug!("decrypting protected = {:?}", protected);
 
         match key {
-            PlainSessionKey::V4 { sym_alg, key } => {
+            PlainSessionKey::V3_4 { sym_alg, ref key } => {
                 ensure!(
                     sym_alg != SymmetricKeyAlgorithm::Plaintext,
                     "session key algorithm cannot be plaintext"
@@ -185,8 +206,8 @@ impl Edata {
                             Some(1),
                             "Version mismatch between key and integrity packet"
                         );
-                        let data = p.decrypt(&key, Some(sym_alg))?;
-                        Message::from_bytes(&data[..])
+                        let data = p.decrypt(key, Some(sym_alg))?;
+                        Self::process_decrypted(&data[..])
                     }
                     Self::SymEncryptedData(p) => {
                         ensure_eq!(
@@ -195,8 +216,8 @@ impl Edata {
                             "Version mismatch between key and integrity packet"
                         );
                         let mut data = p.data().to_vec();
-                        let res = sym_alg.decrypt(&key, &mut data)?;
-                        Message::from_bytes(res)
+                        let res = sym_alg.decrypt(key, &mut data)?;
+                        Self::process_decrypted(res)
                     }
                 }
             }
@@ -213,31 +234,21 @@ impl Edata {
                     bail!("invalid packet combination");
                 }
             },
-            PlainSessionKey::V6 { key } => {
-                match self {
-                    Self::SymEncryptedProtectedData(p) => {
-                        let decrypted_packets = p.decrypt(&key, None)?;
+            PlainSessionKey::V6 { ref key } => match self {
+                Self::SymEncryptedProtectedData(p) => {
+                    ensure_eq!(
+                        self.version(),
+                        Some(2),
+                        "Version mismatch between key and integrity packet"
+                    );
 
-                        let mut messages = Message::from_bytes_many(&decrypted_packets[..]);
-                        // First message is the one we want to return
-                        let Some(message) = messages.next() else {
-                            bail!("no valid message found");
-                        };
-                        let message = message?;
-
-                        // The only other message allowed is a padding packet, which will be skipped
-                        // by the parser, so check that we have only a single message.
-                        if let Some(msg) = messages.next() {
-                            bail!("unexpected message: {:?}", msg);
-                        }
-
-                        Ok(message)
-                    }
-                    Self::SymEncryptedData(_) => {
-                        bail!("invalid packet combination");
-                    }
+                    let decrypted_packets = p.decrypt(key, None)?;
+                    Self::process_decrypted(&decrypted_packets[..])
                 }
-            }
+                Self::SymEncryptedData(_) => {
+                    bail!("invalid packet combination");
+                }
+            },
         }
     }
 }
@@ -322,34 +333,67 @@ impl Message {
         }
     }
 
-    /// Encrypt the message to the list of passed in public keys.
-    pub fn encrypt_to_keys<R: CryptoRng + Rng>(
+    /// Encrypt the message in SEIPDv1 format to a list of public keys `pkeys`.
+    pub fn encrypt_to_keys_seipdv1<R: CryptoRng + Rng>(
         &self,
-        rng: &mut R,
+        mut rng: R,
         alg: SymmetricKeyAlgorithm,
         pkeys: &[&impl PublicKeyTrait],
     ) -> Result<Self> {
         // 1. Generate a session key.
-        let session_key = alg.new_session_key(rng);
+        let session_key = alg.new_session_key(&mut rng);
 
         // 2. Encrypt (pub) the session key, to each PublicKey.
         let esk = pkeys
             .iter()
             .map(|pkey| {
-                let pkes =
-                    PublicKeyEncryptedSessionKey::from_session_key(rng, &session_key, alg, pkey)?;
+                let pkes = PublicKeyEncryptedSessionKey::from_session_key_v3(
+                    &mut rng,
+                    &session_key,
+                    alg,
+                    pkey,
+                )?;
                 Ok(Esk::PublicKeyEncryptedSessionKey(pkes))
             })
             .collect::<Result<_>>()?;
 
         // 3. Encrypt (sym) the data using the session key.
-        self.encrypt_symmetric(rng, esk, alg, session_key)
+        self.encrypt_symmetric_seipdv1(&mut rng, esk, alg, &session_key)
     }
 
-    /// Encrypt the message using the given password.
-    pub fn encrypt_with_password<R, F>(
+    /// Encrypt the message in SEIPDv2 format to a list of public keys `pkeys`.
+    pub fn encrypt_to_keys_seipdv2<R: CryptoRng + Rng>(
         &self,
-        rng: &mut R,
+        mut rng: R,
+        alg: SymmetricKeyAlgorithm,
+        aead: AeadAlgorithm,
+        chunk_size: u8,
+        pkeys: &[&impl PublicKeyTrait],
+    ) -> Result<Self> {
+        // 1. Generate a session key.
+        let session_key = alg.new_session_key(&mut rng);
+
+        // 2. Encrypt (pub) the session key, to each PublicKey.
+        let esk = pkeys
+            .iter()
+            .map(|pkey| {
+                let pkes = PublicKeyEncryptedSessionKey::from_session_key_v6(
+                    &mut rng,
+                    &session_key,
+                    pkey,
+                )?;
+                Ok(Esk::PublicKeyEncryptedSessionKey(pkes))
+            })
+            .collect::<Result<_>>()?;
+
+        // 3. Encrypt (sym) the data using the session key.
+        self.encrypt_symmetric_seipdv2(&mut rng, esk, alg, aead, chunk_size, &session_key)
+    }
+
+    /// Encrypt the message in SEIPDv1 format to a password `msg_pw`.
+    pub fn encrypt_with_password_seipdv1<R, F>(
+        &self,
+        mut rng: R,
         s2k: StringToKey,
         alg: SymmetricKeyAlgorithm,
         msg_pw: F,
@@ -359,10 +403,10 @@ impl Message {
         F: FnOnce() -> String + Clone,
     {
         // 1. Generate a session key.
-        let session_key = alg.new_session_key(rng);
+        let session_key = alg.new_session_key(&mut rng);
 
         // 2. Encrypt (sym) the session key using the provided password.
-        let skesk = Esk::SymKeyEncryptedSessionKey(SymKeyEncryptedSessionKey::encrypt(
+        let skesk = Esk::SymKeyEncryptedSessionKey(SymKeyEncryptedSessionKey::encrypt_v4(
             msg_pw,
             &session_key,
             s2k,
@@ -370,23 +414,82 @@ impl Message {
         )?);
 
         // 3. Encrypt (sym) the data using the session key.
-        self.encrypt_symmetric(rng, vec![skesk], alg, session_key)
+        self.encrypt_symmetric_seipdv1(rng, vec![skesk], alg, &session_key)
     }
 
-    /// Symmetrically encrypts oneself using the provided `session_key`.
-    fn encrypt_symmetric<R: CryptoRng + Rng>(
+    /// Encrypt the message in SEIPDv2 format to a password `msg_pw`.
+    pub fn encrypt_with_password_seipdv2<R, F>(
         &self,
-        rng: &mut R,
+        mut rng: R,
+        s2k: StringToKey,
+        alg: SymmetricKeyAlgorithm,
+        aead: AeadAlgorithm,
+        chunk_size: u8,
+        msg_pw: F,
+    ) -> Result<Self>
+    where
+        R: Rng + CryptoRng,
+        F: FnOnce() -> String + Clone,
+    {
+        // 1. Generate a session key.
+        let session_key = alg.new_session_key(&mut rng);
+
+        // 2. Encrypt (sym) the session key using the provided password.
+        let skesk = Esk::SymKeyEncryptedSessionKey(SymKeyEncryptedSessionKey::encrypt_v6(
+            &mut rng,
+            msg_pw,
+            &session_key,
+            s2k,
+            alg,
+            aead,
+        )?);
+
+        // 3. Encrypt (sym) the data using the session key.
+        self.encrypt_symmetric_seipdv2(rng, vec![skesk], alg, aead, chunk_size, &session_key)
+    }
+
+    /// Symmetrically encrypt this Message in SEIPDv1 format using the provided `session_key`.
+    ///
+    /// This function assumes that it is only called with Esk that are legal to use with SEIPDv1.
+    fn encrypt_symmetric_seipdv1<R: CryptoRng + Rng>(
+        &self,
+        rng: R,
         esk: Vec<Esk>,
         alg: SymmetricKeyAlgorithm,
-        session_key: Vec<u8>,
+        session_key: &[u8],
     ) -> Result<Self> {
         let data = self.to_bytes()?;
 
-        let edata = Edata::SymEncryptedProtectedData(SymEncryptedProtectedData::encrypt_with_rng(
+        let edata = Edata::SymEncryptedProtectedData(SymEncryptedProtectedData::encrypt_seipdv1(
             rng,
             alg,
-            &session_key,
+            session_key,
+            &data,
+        )?);
+
+        Ok(Message::Encrypted { esk, edata })
+    }
+
+    /// Symmetrically encrypt this Message in SEIPDv2 format using the provided `session_key`.
+    ///
+    /// This function assumes that it is only called with Esk that are legal to use with SEIPDv2.
+    fn encrypt_symmetric_seipdv2<R: CryptoRng + Rng>(
+        &self,
+        rng: R,
+        esk: Vec<Esk>,
+        alg: SymmetricKeyAlgorithm,
+        aead: AeadAlgorithm,
+        chunk_size: u8,
+        session_key: &[u8],
+    ) -> Result<Self> {
+        let data = self.to_bytes()?;
+
+        let edata = Edata::SymEncryptedProtectedData(SymEncryptedProtectedData::encrypt_seipdv2(
+            rng,
+            alg,
+            aead,
+            chunk_size,
+            session_key,
             &data,
         )?);
 
@@ -394,22 +497,22 @@ impl Message {
     }
 
     /// Sign this message using the provided key.
-    pub fn sign<F>(
+    pub fn sign<R, F>(
         self,
+        rng: R,
         key: &impl SecretKeyTrait,
         key_pw: F,
         hash_algorithm: HashAlgorithm,
     ) -> Result<Self>
     where
+        R: CryptoRng + Rng,
         F: FnOnce() -> String,
     {
         let key_id = key.key_id();
         let algorithm = key.algorithm();
+
         let hashed_subpackets = vec![
-            Subpacket::regular(SubpacketData::IssuerFingerprint(
-                KeyVersion::V4,
-                SmallVec::from_slice(&key.fingerprint()),
-            )),
+            Subpacket::regular(SubpacketData::IssuerFingerprint(key.fingerprint())),
             Subpacket::regular(SubpacketData::SignatureCreationTime(
                 chrono::Utc::now().trunc_subsecs(0),
             )),
@@ -424,34 +527,51 @@ impl Message {
                     SignatureType::Text
                 };
 
-                let signature_config = SignatureConfig::new_v4(
-                    Default::default(),
-                    typ,
-                    algorithm,
-                    hash_algorithm,
-                    hashed_subpackets,
-                    unhashed_subpackets,
-                );
-                (typ, signature_config.sign(key, key_pw, l.data())?)
+                let mut config = match key.version() {
+                    KeyVersion::V4 => SignatureConfig::v4(typ, algorithm, hash_algorithm),
+                    KeyVersion::V6 => SignatureConfig::v6(rng, typ, algorithm, hash_algorithm)?,
+                    v => bail!("unsupported key version {:?}", v),
+                };
+                config.hashed_subpackets = hashed_subpackets;
+                config.unhashed_subpackets = unhashed_subpackets;
+
+                (typ, config.sign(key, key_pw, l.data())?)
             }
             _ => {
                 let typ = SignatureType::Binary;
-                let signature_config = SignatureConfig::new_v4(
-                    Default::default(),
-                    typ,
-                    algorithm,
-                    hash_algorithm,
-                    hashed_subpackets,
-                    unhashed_subpackets,
-                );
+
+                let mut config = match key.version() {
+                    KeyVersion::V4 => SignatureConfig::v4(typ, algorithm, hash_algorithm),
+                    KeyVersion::V6 => SignatureConfig::v6(rng, typ, algorithm, hash_algorithm)?,
+                    v => bail!("unsupported key version {:?}", v),
+                };
+                config.hashed_subpackets = hashed_subpackets;
+                config.unhashed_subpackets = unhashed_subpackets;
 
                 let data = self.to_bytes()?;
-                let signature = signature_config.sign(key, key_pw, &data[..])?;
+                let signature = config.sign(key, key_pw, &data[..])?;
 
                 (typ, signature)
             }
         };
-        let ops = OnePassSignature::from_details(typ, hash_algorithm, algorithm, key_id);
+
+        let ops = match key.version() {
+            KeyVersion::V4 => OnePassSignature::v3(typ, hash_algorithm, algorithm, key_id),
+            KeyVersion::V6 => {
+                let SignatureVersionSpecific::V6 { ref salt } = signature.config.version_specific
+                else {
+                    // This should never happen
+                    bail!("Inconsistent Signature and OnePassSignature version")
+                };
+
+                let Fingerprint::V6(fp) = key.fingerprint() else {
+                    bail!("Inconsistent Signature and Fingerprint version")
+                };
+
+                OnePassSignature::v6(typ, hash_algorithm, algorithm, salt.clone(), fp)
+            }
+            v => bail!("Unsupported key version {:?}", v),
+        };
 
         Ok(Message::Signed {
             message: Some(Box::new(self)),
@@ -514,20 +634,6 @@ impl Message {
         }
     }
 
-    /// Returns a list of [KeyId]s that the message is encrypted to. For non encrypted messages this list is empty.
-    pub fn get_recipients(&self) -> Vec<&KeyId> {
-        match self {
-            Message::Encrypted { esk, .. } => esk
-                .iter()
-                .filter_map(|e| match e {
-                    Esk::PublicKeyEncryptedSessionKey(k) => Some(k.id()),
-                    _ => None,
-                })
-                .collect(),
-            _ => Vec::new(),
-        }
-    }
-
     /// Decrypt the message using the given key.
     /// Returns a message decrypter, and a list of [KeyId]s that are valid recipients of this message.
     pub fn decrypt<G>(&self, key_pw: G, keys: &[&SignedSecretKey]) -> Result<(Message, Vec<KeyId>)>
@@ -561,13 +667,13 @@ impl Message {
                                 "{:?}",
                                 key.secret_subkeys
                                     .iter()
-                                    .map(KeyTrait::key_id)
+                                    .map(PublicKeyTrait::key_id)
                                     .collect::<Vec<_>>()
                             );
 
-                            // find the key with the matching key id
+                            // find the matching key or subkey
 
-                            if &key.primary_key.key_id() == esk_packet.id() {
+                            if esk_packet.match_identity(&key.primary_key) {
                                 encoding_key = Some(&key.primary_key);
                             }
 
@@ -575,7 +681,7 @@ impl Message {
                                 encoding_subkey = key
                                     .secret_subkeys
                                     .iter()
-                                    .find(|&subkey| &subkey.key_id() == esk_packet.id());
+                                    .find(|&subkey| esk_packet.match_identity(&subkey));
                             }
 
                             if encoding_key.is_some() || encoding_subkey.is_some() {
@@ -594,16 +700,24 @@ impl Message {
 
                 let session_keys = valid_keys
                     .iter()
-                    .map(|(packet, encoding_key, encoding_subkey)| {
+                    .map(|(pkesk, encoding_key, encoding_subkey)| {
+                        let typ = match pkesk.version() {
+                            PkeskVersion::V3 => EskType::V3_4,
+                            PkeskVersion::V6 => EskType::V6,
+                            PkeskVersion::Other(v) => {
+                                unimplemented_err!("Unexpected PKESK version {}", v)
+                            }
+                        };
+
                         if let Some(ek) = encoding_key {
                             Ok((
                                 ek.key_id(),
-                                decrypt_session_key(ek, key_pw.clone(), packet.mpis())?,
+                                decrypt_session_key(ek, key_pw.clone(), pkesk.values()?, typ)?,
                             ))
                         } else if let Some(ek) = encoding_subkey {
                             Ok((
                                 ek.key_id(),
-                                decrypt_session_key(ek, key_pw.clone(), packet.mpis())?,
+                                decrypt_session_key(ek, key_pw.clone(), pkesk.values()?, typ)?,
                             ))
                         } else {
                             unreachable!("either a key or a subkey were found");
@@ -786,10 +900,15 @@ impl<'a> From<Option<&'a armor::Headers>> for ArmorOptions<'a> {
 mod tests {
     #![allow(clippy::unwrap_used)]
 
-    use super::*;
-    use rand::thread_rng;
     use std::fs;
 
+    use rand::SeedableRng;
+    use rand_chacha::ChaCha8Rng;
+
+    use super::*;
+    use crate::cleartext::CleartextSignedMessage;
+    use crate::SignedPublicKey;
+
     #[test]
     fn test_compression_zlib() {
         let lit_msg = Message::new_literal("hello-zlib.txt", "hello world");
@@ -823,11 +942,9 @@ mod tests {
     }
 
     #[test]
-    fn test_rsa_encryption() {
-        use rand::SeedableRng;
-
+    fn test_rsa_encryption_seipdv1() {
         let (skey, _headers) = SignedSecretKey::from_armor_single(
-            fs::File::open("./tests/opengpg-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
+            fs::File::open("./tests/openpgp-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
                 .unwrap(),
         )
         .unwrap();
@@ -842,10 +959,57 @@ mod tests {
 
         // Encrypt and test that rng is the only source of randomness.
         let encrypted = compressed_msg
-            .encrypt_to_keys(&mut rng, SymmetricKeyAlgorithm::AES128, &[&pkey][..])
+            .encrypt_to_keys_seipdv1(&mut rng, SymmetricKeyAlgorithm::AES128, &[&pkey][..])
             .unwrap();
         let encrypted2 = compressed_msg
-            .encrypt_to_keys(&mut rng2, SymmetricKeyAlgorithm::AES128, &[&pkey][..])
+            .encrypt_to_keys_seipdv1(&mut rng2, SymmetricKeyAlgorithm::AES128, &[&pkey][..])
+            .unwrap();
+        assert_eq!(encrypted, encrypted2);
+
+        let armored = encrypted.to_armored_bytes(None.into()).unwrap();
+        // fs::write("./message-rsa.asc", &armored).unwrap();
+
+        let parsed = Message::from_armor_single(&armored[..]).unwrap().0;
+
+        let decrypted = parsed.decrypt(|| "test".into(), &[&skey]).unwrap().0;
+
+        assert_eq!(compressed_msg, decrypted);
+    }
+
+    #[test]
+    fn test_rsa_encryption_seipdv2() {
+        let (skey, _headers) = SignedSecretKey::from_armor_single(
+            fs::File::open("./tests/openpgp-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
+                .unwrap(),
+        )
+        .unwrap();
+
+        // subkey[0] is the encryption key
+        let pkey = skey.secret_subkeys[0].public_key();
+        let mut rng = rand::rngs::StdRng::seed_from_u64(100);
+        let mut rng2 = rand::rngs::StdRng::seed_from_u64(100);
+
+        let lit_msg = Message::new_literal("hello.txt", "hello world\n");
+        let compressed_msg = lit_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
+
+        // Encrypt and test that rng is the only source of randomness.
+        let encrypted = compressed_msg
+            .encrypt_to_keys_seipdv2(
+                &mut rng,
+                SymmetricKeyAlgorithm::AES128,
+                AeadAlgorithm::Ocb,
+                0x06,
+                &[&pkey][..],
+            )
+            .unwrap();
+        let encrypted2 = compressed_msg
+            .encrypt_to_keys_seipdv2(
+                &mut rng2,
+                SymmetricKeyAlgorithm::AES128,
+                AeadAlgorithm::Ocb,
+                0x06,
+                &[&pkey][..],
+            )
             .unwrap();
         assert_eq!(encrypted, encrypted2);
 
@@ -860,7 +1024,7 @@ mod tests {
     }
 
     #[test]
-    fn test_x25519_encryption() {
+    fn test_x25519_encryption_seipdv1() {
         let (skey, _headers) = SignedSecretKey::from_armor_single(
             fs::File::open("./tests/autocrypt/alice@autocrypt.example.sec.asc").unwrap(),
         )
@@ -868,17 +1032,17 @@ mod tests {
 
         // subkey[0] is the encryption key
         let pkey = skey.secret_subkeys[0].public_key();
-        let mut rng = thread_rng();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal("hello.txt", "hello world\n");
         let compressed_msg = lit_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
         for _ in 0..1000 {
             let encrypted = compressed_msg
-                .encrypt_to_keys(&mut rng, SymmetricKeyAlgorithm::AES128, &[&pkey][..])
+                .encrypt_to_keys_seipdv1(&mut rng, SymmetricKeyAlgorithm::AES128, &[&pkey][..])
                 .unwrap();
 
             let armored = encrypted.to_armored_bytes(None.into()).unwrap();
-            fs::write("./message-x25519.asc", &armored).unwrap();
+            // fs::write("./message-x25519.asc", &armored).unwrap();
 
             let parsed = Message::from_armor_single(&armored[..]).unwrap().0;
 
@@ -889,10 +1053,48 @@ mod tests {
     }
 
     #[test]
-    fn test_password_encryption() {
+    fn test_x25519_encryption_seipdv2() {
+        let (skey, _headers) = SignedSecretKey::from_armor_single(
+            fs::File::open("./tests/autocrypt/alice@autocrypt.example.sec.asc").unwrap(),
+        )
+        .unwrap();
+
+        // subkey[0] is the encryption key
+        let pkey = skey.secret_subkeys[0].public_key();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
+
+        let lit_msg = Message::new_literal("hello.txt", "hello world\n");
+        let compressed_msg = lit_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
+
+        for aead in [AeadAlgorithm::Ocb, AeadAlgorithm::Eax, AeadAlgorithm::Gcm] {
+            for sym in [
+                SymmetricKeyAlgorithm::AES128,
+                SymmetricKeyAlgorithm::AES192,
+                SymmetricKeyAlgorithm::AES256,
+            ] {
+                for _ in 0..1000 {
+                    let encrypted = compressed_msg
+                        .encrypt_to_keys_seipdv2(&mut rng, sym, aead, 0x06, &[&pkey][..])
+                        .unwrap();
+
+                    let armored = encrypted.to_armored_bytes(None.into()).unwrap();
+                    fs::write("./message-x25519.asc", &armored).unwrap();
+
+                    let parsed = Message::from_armor_single(&armored[..]).unwrap().0;
+
+                    let decrypted = parsed.decrypt(|| "".into(), &[&skey]).unwrap().0;
+
+                    assert_eq!(compressed_msg, decrypted);
+                }
+            }
+        }
+    }
+
+    #[test]
+    fn test_password_encryption_seipdv1() {
         let _ = pretty_env_logger::try_init();
 
-        let mut rng = thread_rng();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal("hello.txt", "hello world\n");
         let compressed_msg = lit_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
@@ -900,13 +1102,13 @@ mod tests {
         let s2k = StringToKey::new_default(&mut rng);
 
         let encrypted = compressed_msg
-            .encrypt_with_password(&mut rng, s2k, SymmetricKeyAlgorithm::AES128, || {
+            .encrypt_with_password_seipdv1(&mut rng, s2k, SymmetricKeyAlgorithm::AES128, || {
                 "secret".into()
             })
             .unwrap();
 
         let armored = encrypted.to_armored_bytes(None.into()).unwrap();
-        fs::write("./message-password.asc", &armored).unwrap();
+        // fs::write("./message-password.asc", &armored).unwrap();
 
         let parsed = Message::from_armor_single(&armored[..]).unwrap().0;
 
@@ -915,6 +1117,41 @@ mod tests {
         assert_eq!(compressed_msg, decrypted);
     }
 
+    #[test]
+    fn test_password_encryption_seipdv2() {
+        let _ = pretty_env_logger::try_init();
+
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
+
+        let lit_msg = Message::new_literal("hello.txt", "hello world\n");
+        let compressed_msg = lit_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
+
+        for aead in [AeadAlgorithm::Ocb, AeadAlgorithm::Eax, AeadAlgorithm::Gcm] {
+            for sym in [
+                SymmetricKeyAlgorithm::AES128,
+                SymmetricKeyAlgorithm::AES192,
+                SymmetricKeyAlgorithm::AES256,
+            ] {
+                let s2k = StringToKey::new_default(&mut rng);
+
+                let encrypted = compressed_msg
+                    .encrypt_with_password_seipdv2(&mut rng, s2k, sym, aead, 0x06, || {
+                        "secret".into()
+                    })
+                    .unwrap();
+
+                let armored = encrypted.to_armored_bytes(None.into()).unwrap();
+                fs::write("./message-password.asc", &armored).unwrap();
+
+                let parsed = Message::from_armor_single(&armored[..]).unwrap().0;
+
+                let decrypted = parsed.decrypt_with_password(|| "secret".into()).unwrap();
+
+                assert_eq!(compressed_msg, decrypted);
+            }
+        }
+    }
+
     #[test]
     fn test_no_plaintext_decryption() {
         // Invalid message "encrypted" with plaintext algorithm.
@@ -972,16 +1209,17 @@ mod tests {
         .unwrap();
 
         let pkey = skey.public_key();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal("hello.txt", "hello world\n");
         assert!(lit_msg.verify(&pkey).is_err()); // Unsigned message shouldn't verify
 
         let signed_msg = lit_msg
-            .sign(&skey, || "".into(), HashAlgorithm::SHA2_256)
+            .sign(&mut rng, &skey, || "".into(), HashAlgorithm::SHA2_256)
             .unwrap();
 
         let armored = signed_msg.to_armored_bytes(None.into()).unwrap();
-        fs::write("./message-string-signed-x25519.asc", &armored).unwrap();
+        // fs::write("./message-string-signed-x25519.asc", &armored).unwrap();
 
         signed_msg.verify(&pkey).unwrap();
 
@@ -997,14 +1235,15 @@ mod tests {
         .unwrap();
 
         let pkey = skey.public_key();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal_bytes("hello.txt", &b"hello world\n"[..]);
         let signed_msg = lit_msg
-            .sign(&skey, || "".into(), HashAlgorithm::SHA2_256)
+            .sign(&mut rng, &skey, || "".into(), HashAlgorithm::SHA2_256)
             .unwrap();
 
         let armored = signed_msg.to_armored_bytes(None.into()).unwrap();
-        fs::write("./message-bytes-signed-x25519.asc", &armored).unwrap();
+        // fs::write("./message-bytes-signed-x25519.asc", &armored).unwrap();
 
         signed_msg.verify(&pkey).unwrap();
 
@@ -1020,15 +1259,16 @@ mod tests {
         .unwrap();
 
         let pkey = skey.public_key();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal_bytes("hello.txt", &b"hello world\n"[..]);
         let signed_msg = lit_msg
-            .sign(&skey, || "".into(), HashAlgorithm::SHA2_256)
+            .sign(&mut rng, &skey, || "".into(), HashAlgorithm::SHA2_256)
             .unwrap();
         let compressed_msg = signed_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
 
         let armored = compressed_msg.to_armored_bytes(None.into()).unwrap();
-        fs::write("./message-bytes-compressed-signed-x25519.asc", &armored).unwrap();
+        // fs::write("./message-bytes-compressed-signed-x25519.asc", &armored).unwrap();
 
         signed_msg.verify(&pkey).unwrap();
 
@@ -1041,21 +1281,22 @@ mod tests {
         for _ in 0..100 {
             let (skey, _headers) = SignedSecretKey::from_armor_single(
                 fs::File::open(
-                    "./tests/opengpg-interop/testcases/messages/gnupg-v1-001-decrypt.asc",
+                    "./tests/openpgp-interop/testcases/messages/gnupg-v1-001-decrypt.asc",
                 )
                 .unwrap(),
             )
             .unwrap();
 
             let pkey = skey.public_key();
+            let mut rng = ChaCha8Rng::seed_from_u64(0);
 
             let lit_msg = Message::new_literal("hello.txt", "hello world\n");
             let signed_msg = lit_msg
-                .sign(&skey, || "test".into(), HashAlgorithm::SHA2_256)
+                .sign(&mut rng, &skey, || "test".into(), HashAlgorithm::SHA2_256)
                 .unwrap();
 
             let armored = signed_msg.to_armored_bytes(None.into()).unwrap();
-            fs::write("./message-string-signed-rsa.asc", &armored).unwrap();
+            // fs::write("./message-string-signed-rsa.asc", &armored).unwrap();
 
             signed_msg.verify(&pkey).unwrap();
 
@@ -1067,20 +1308,21 @@ mod tests {
     #[test]
     fn test_rsa_signing_bytes() {
         let (skey, _headers) = SignedSecretKey::from_armor_single(
-            fs::File::open("./tests/opengpg-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
+            fs::File::open("./tests/openpgp-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
                 .unwrap(),
         )
         .unwrap();
 
         let pkey = skey.public_key();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal_bytes("hello.txt", &b"hello world\n"[..]);
         let signed_msg = lit_msg
-            .sign(&skey, || "test".into(), HashAlgorithm::SHA2_256)
+            .sign(&mut rng, &skey, || "test".into(), HashAlgorithm::SHA2_256)
             .unwrap();
 
         let armored = signed_msg.to_armored_bytes(None.into()).unwrap();
-        fs::write("./message-bytes-signed-rsa.asc", &armored).unwrap();
+        // fs::write("./message-bytes-signed-rsa.asc", &armored).unwrap();
 
         signed_msg.verify(&pkey).unwrap();
 
@@ -1091,21 +1333,22 @@ mod tests {
     #[test]
     fn test_rsa_signing_bytes_compressed() {
         let (skey, _headers) = SignedSecretKey::from_armor_single(
-            fs::File::open("./tests/opengpg-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
+            fs::File::open("./tests/openpgp-interop/testcases/messages/gnupg-v1-001-decrypt.asc")
                 .unwrap(),
         )
         .unwrap();
 
         let pkey = skey.public_key();
+        let mut rng = ChaCha8Rng::seed_from_u64(0);
 
         let lit_msg = Message::new_literal_bytes("hello.txt", &b"hello world\n"[..]);
         let signed_msg = lit_msg
-            .sign(&skey, || "test".into(), HashAlgorithm::SHA2_256)
+            .sign(&mut rng, &skey, || "test".into(), HashAlgorithm::SHA2_256)
             .unwrap();
 
         let compressed_msg = signed_msg.compress(CompressionAlgorithm::ZLIB).unwrap();
         let armored = compressed_msg.to_armored_bytes(None.into()).unwrap();
-        fs::write("./message-bytes-compressed-signed-rsa.asc", &armored).unwrap();
+        // fs::write("./message-bytes-compressed-signed-rsa.asc", &armored).unwrap();
 
         signed_msg.verify(&pkey).unwrap();
 
@@ -1122,7 +1365,7 @@ mod tests {
         // The literal data packet (which is in binary mode) contains the output of:
         // echo -en "foo\nbar\r\nbaz"
         //
-        // RFC 4880 mandates that the hash for signature type 0x01 has to be calculated over normalized line endings,
+        // RFC 9580 mandates that the hash for signature type 0x01 has to be calculated over normalized line endings,
         // so the hash for this message is calculated over "foo\r\nbar\r\nbaz".
         //
         // So it must also be verified against a hash digest over this normalized format.
@@ -1169,4 +1412,116 @@ mod tests {
         assert!(msg.get_content().is_err());
         assert!(msg.verify(&pkey).is_err());
     }
+
+    // Sample Version 6 Certificate (Transferable Public Key)
+    // https://www.rfc-editor.org/rfc/rfc9580.html#name-sample-version-6-certificat
+    const ANNEX_A_3: &str = "-----BEGIN PGP PUBLIC KEY BLOCK-----
+
+xioGY4d/4xsAAAAg+U2nu0jWCmHlZ3BqZYfQMxmZu52JGggkLq2EVD34laPCsQYf
+GwoAAABCBYJjh3/jAwsJBwUVCg4IDAIWAAKbAwIeCSIhBssYbE8GCaaX5NUt+mxy
+KwwfHifBilZwj2Ul7Ce62azJBScJAgcCAAAAAK0oIBA+LX0ifsDm185Ecds2v8lw
+gyU2kCcUmKfvBXbAf6rhRYWzuQOwEn7E/aLwIwRaLsdry0+VcallHhSu4RN6HWaE
+QsiPlR4zxP/TP7mhfVEe7XWPxtnMUMtf15OyA51YBM4qBmOHf+MZAAAAIIaTJINn
++eUBXbki+PSAld2nhJh/LVmFsS+60WyvXkQ1wpsGGBsKAAAALAWCY4d/4wKbDCIh
+BssYbE8GCaaX5NUt+mxyKwwfHifBilZwj2Ul7Ce62azJAAAAAAQBIKbpGG2dWTX8
+j+VjFM21J0hqWlEg+bdiojWnKfA5AQpWUWtnNwDEM0g12vYxoWM8Y81W+bHBw805
+I8kWVkXU6vFOi+HWvv/ira7ofJu16NnoUkhclkUrk0mXubZvyl4GBg==
+-----END PGP PUBLIC KEY BLOCK-----";
+
+    // Sample Version 6 Secret Key (Transferable Secret Key)
+    // https://www.rfc-editor.org/rfc/rfc9580.html#name-sample-version-6-secret-key
+    const ANNEX_A_4: &str = "-----BEGIN PGP PRIVATE KEY BLOCK-----
+
+xUsGY4d/4xsAAAAg+U2nu0jWCmHlZ3BqZYfQMxmZu52JGggkLq2EVD34laMAGXKB
+exK+cH6NX1hs5hNhIB00TrJmosgv3mg1ditlsLfCsQYfGwoAAABCBYJjh3/jAwsJ
+BwUVCg4IDAIWAAKbAwIeCSIhBssYbE8GCaaX5NUt+mxyKwwfHifBilZwj2Ul7Ce6
+2azJBScJAgcCAAAAAK0oIBA+LX0ifsDm185Ecds2v8lwgyU2kCcUmKfvBXbAf6rh
+RYWzuQOwEn7E/aLwIwRaLsdry0+VcallHhSu4RN6HWaEQsiPlR4zxP/TP7mhfVEe
+7XWPxtnMUMtf15OyA51YBMdLBmOHf+MZAAAAIIaTJINn+eUBXbki+PSAld2nhJh/
+LVmFsS+60WyvXkQ1AE1gCk95TUR3XFeibg/u/tVY6a//1q0NWC1X+yui3O24wpsG
+GBsKAAAALAWCY4d/4wKbDCIhBssYbE8GCaaX5NUt+mxyKwwfHifBilZwj2Ul7Ce6
+2azJAAAAAAQBIKbpGG2dWTX8j+VjFM21J0hqWlEg+bdiojWnKfA5AQpWUWtnNwDE
+M0g12vYxoWM8Y81W+bHBw805I8kWVkXU6vFOi+HWvv/ira7ofJu16NnoUkhclkUr
+k0mXubZvyl4GBg==
+-----END PGP PRIVATE KEY BLOCK-----";
+
+    /// Verify Cleartext Signed Message
+    ///
+    /// Test data from RFC 9580, see
+    /// https://www.rfc-editor.org/rfc/rfc9580.html#name-sample-cleartext-signed-mes
+    #[test]
+    fn test_v6_annex_a_6() {
+        let (ssk, _) = SignedPublicKey::from_string(ANNEX_A_3).expect("SSK from armor");
+
+        let msg = "-----BEGIN PGP SIGNED MESSAGE-----
+
+What we need from the grocery store:
+
+- - tofu
+- - vegetables
+- - noodles
+
+-----BEGIN PGP SIGNATURE-----
+
+wpgGARsKAAAAKQWCY5ijYyIhBssYbE8GCaaX5NUt+mxyKwwfHifBilZwj2Ul7Ce6
+2azJAAAAAGk2IHZJX1AhiJD39eLuPBgiUU9wUA9VHYblySHkBONKU/usJ9BvuAqo
+/FvLFuGWMbKAdA+epq7V4HOtAPlBWmU8QOd6aud+aSunHQaaEJ+iTFjP2OMW0KBr
+NK2ay45cX1IVAQ==
+-----END PGP SIGNATURE-----";
+
+        let (msg, _) = CleartextSignedMessage::from_string(msg).unwrap();
+
+        msg.verify(&ssk).expect("verify");
+    }
+
+    /// Verify Inline Signed Message
+    ///
+    /// Test data from RFC 9580, see
+    /// https://www.rfc-editor.org/rfc/rfc9580.html#name-sample-inline-signed-messag
+    #[test]
+    fn test_v6_annex_a_7() {
+        let (ssk, _) = SignedPublicKey::from_string(ANNEX_A_3).expect("SSK from armor");
+
+        let msg = "-----BEGIN PGP MESSAGE-----
+
+xEYGAQobIHZJX1AhiJD39eLuPBgiUU9wUA9VHYblySHkBONKU/usyxhsTwYJppfk
+1S36bHIrDB8eJ8GKVnCPZSXsJ7rZrMkBy0p1AAAAAABXaGF0IHdlIG5lZWQgZnJv
+bSB0aGUgZ3JvY2VyeSBzdG9yZToKCi0gdG9mdQotIHZlZ2V0YWJsZXMKLSBub29k
+bGVzCsKYBgEbCgAAACkFgmOYo2MiIQbLGGxPBgmml+TVLfpscisMHx4nwYpWcI9l
+JewnutmsyQAAAABpNiB2SV9QIYiQ9/Xi7jwYIlFPcFAPVR2G5ckh5ATjSlP7rCfQ
+b7gKqPxbyxbhljGygHQPnqau1eBzrQD5QVplPEDnemrnfmkrpx0GmhCfokxYz9jj
+FtCgazStmsuOXF9SFQE=
+-----END PGP MESSAGE-----";
+
+        let (msg, _) = Message::from_string(msg).unwrap();
+
+        msg.verify(&ssk).expect("verify");
+    }
+
+    /// Decrypt an X25519-AEAD-OCB Encrypted Packet Sequence
+    ///
+    /// Test data from RFC 9580, see
+    /// https://www.rfc-editor.org/rfc/rfc9580.html#name-sample-x25519-aead-ocb-encr
+    #[test]
+    fn test_v6_annex_a_8() {
+        let (ssk, _) = SignedSecretKey::from_string(ANNEX_A_4).expect("SSK from armor");
+
+        // A.8. Sample X25519-AEAD-OCB Decryption
+        let msg = "-----BEGIN PGP MESSAGE-----
+
+wV0GIQYSyD8ecG9jCP4VGkF3Q6HwM3kOk+mXhIjR2zeNqZMIhRmHzxjV8bU/gXzO
+WgBM85PMiVi93AZfJfhK9QmxfdNnZBjeo1VDeVZheQHgaVf7yopqR6W1FT6NOrfS
+aQIHAgZhZBZTW+CwcW1g4FKlbExAf56zaw76/prQoN+bAzxpohup69LA7JW/Vp0l
+yZnuSj3hcFj0DfqLTGgr4/u717J+sPWbtQBfgMfG9AOIwwrUBqsFE9zW+f1zdlYo
+bhF30A+IitsxxA==
+-----END PGP MESSAGE-----";
+
+        let (message, _) = Message::from_string(msg).expect("ok");
+        let (dec, _) = message.decrypt(String::default, &[&ssk]).expect("decrypt");
+
+        let decrypted =
+            String::from_utf8(dec.get_literal().expect("literal").data().to_vec()).expect("utf8");
+
+        assert_eq!(&decrypted, "Hello, world!");
+    }
 }
