@@ -1,8 +1,7 @@
 //! # Cryptography module
 
-use crate::types::{Mpi, PublicParams};
-
 use self::hash::HashAlgorithm;
+use crate::types::PublicParams;
 
 pub mod aead;
 pub mod aes_kw;
@@ -16,9 +15,13 @@ pub mod hash;
 pub mod public_key;
 pub mod rsa;
 pub mod sym;
+pub mod x25519;
+pub mod x448;
 
 pub trait Decryptor {
-    fn decrypt(&self, mpis: &[Mpi], fingerprint: &[u8]) -> crate::errors::Result<Vec<u8>>;
+    type EncryptionFields<'a>;
+
+    fn decrypt(&self, data: Self::EncryptionFields<'_>) -> crate::errors::Result<Vec<u8>>;
 }
 
 pub trait Signer {
