@@ -6,9 +6,11 @@ use cfb_mode::cipher::{AsyncStreamCipher, KeyIvInit};
 use cfb_mode::{BufDecryptor, BufEncryptor, Decryptor, Encryptor};
 use des::TdesEde3;
 use idea::Idea;
+use log::debug;
 use num_enum::{FromPrimitive, IntoPrimitive};
-use rand::{thread_rng, CryptoRng, Rng};
+use rand::{CryptoRng, Rng};
 use twofish::Twofish;
+use zeroize::Zeroizing;
 
 use crate::errors::{Error, Result};
 
@@ -17,10 +19,10 @@ macro_rules! decrypt {
         let mut mode = BufDecryptor::<$mode>::new_from_slices($key, $iv)?;
         mode.decrypt($prefix);
 
-        // We do not do "quick check" here.
-        // See "Security Considerations" section
-        // in <https://tools.ietf.org/html/rfc4880#page-84>
-        // and paper <https://eprint.iacr.org/2005/033>
+        // We do not do use "quick check" here.
+        // See the "Security Considerations" section
+        // in <https://www.rfc-editor.org/rfc/rfc9580.html#name-risks-of-a-quick-check-orac>
+        // and the paper <https://eprint.iacr.org/2005/033>
         // for details.
 
         if $resync {
@@ -63,7 +65,8 @@ macro_rules! encrypt_regular {
     }};
 }
 
-/// Available [symmetric key algorithms](https://tools.ietf.org/html/rfc4880#section-9.2).
+/// Available symmetric key algorithms.
+/// Ref: <https://www.rfc-editor.org/rfc/rfc9580.html#name-symmetric-key-algorithms>
 #[derive(Debug, PartialEq, Eq, Copy, Clone, FromPrimitive, IntoPrimitive)]
 #[repr(u8)]
 pub enum SymmetricKeyAlgorithm {
@@ -108,7 +111,7 @@ impl zeroize::DefaultIsZeroes for SymmetricKeyAlgorithm {}
 
 impl SymmetricKeyAlgorithm {
     /// The size of a single block in bytes.
-    /// Based on https://github.com/gpg/libgcrypt/blob/master/cipher
+    /// Based on <https://github.com/gpg/libgcrypt/blob/master/cipher>
     pub fn block_size(self) -> usize {
         match self {
             SymmetricKeyAlgorithm::Plaintext => 0,
@@ -128,8 +131,8 @@ impl SymmetricKeyAlgorithm {
     }
 
     /// The size of a single block in bytes.
-    /// Based on https://github.com/gpg/libgcrypt/blob/master/cipher
-    pub fn key_size(self) -> usize {
+    /// Based on <https://github.com/gpg/libgcrypt/blob/master/cipher>
+    pub const fn key_size(self) -> usize {
         match self {
             SymmetricKeyAlgorithm::Plaintext => 0,
             SymmetricKeyAlgorithm::IDEA => 16,
@@ -226,7 +229,9 @@ impl SymmetricKeyAlgorithm {
 
         {
             match self {
-                SymmetricKeyAlgorithm::Plaintext => {}
+                SymmetricKeyAlgorithm::Plaintext => {
+                    bail!("'Plaintext' is not a legal cipher for encrypted data")
+                }
                 SymmetricKeyAlgorithm::IDEA => decrypt!(
                     Idea,
                     key,
@@ -347,7 +352,9 @@ impl SymmetricKeyAlgorithm {
         ciphertext: &mut [u8],
     ) -> Result<()> {
         match self {
-            SymmetricKeyAlgorithm::Plaintext => {}
+            SymmetricKeyAlgorithm::Plaintext => {
+                bail!("'Plaintext' is not a legal cipher for encrypted data")
+            }
             SymmetricKeyAlgorithm::IDEA => {
                 decrypt_regular!(Idea, key, iv_vec, ciphertext)
             }
@@ -389,9 +396,9 @@ impl SymmetricKeyAlgorithm {
 
     /// Encrypt the data using CFB mode, without padding. Overwrites the input.
     /// Uses an IV of all zeroes, as specified in the openpgp cfb mode.
-    pub fn encrypt_with_rng<R: CryptoRng + Rng>(
+    pub fn encrypt<R: CryptoRng + Rng>(
         self,
-        rng: &mut R,
+        mut rng: R,
         key: &[u8],
         plaintext: &[u8],
     ) -> Result<Vec<u8>> {
@@ -420,17 +427,9 @@ impl SymmetricKeyAlgorithm {
         Ok(ciphertext)
     }
 
-    /// Same as [`encrypt_with_rng`], but uses [`rand::thread_rng`] for RNG.
-    ///
-    /// [`encrypt_with_rng`]: SymmetricKeyAlgorithm::encrypt_with_rng
-    /// [`rand::thread_rng`]: rand::thread_rng
-    pub fn encrypt(self, key: &[u8], plaintext: &[u8]) -> Result<Vec<u8>> {
-        self.encrypt_with_rng(&mut thread_rng(), key, plaintext)
-    }
-
-    pub fn encrypt_protected_with_rng<R: CryptoRng + Rng>(
+    pub fn encrypt_protected<R: CryptoRng + Rng>(
         self,
-        rng: &mut R,
+        mut rng: R,
         key: &[u8],
         plaintext: &[u8],
     ) -> Result<Vec<u8>> {
@@ -473,10 +472,6 @@ impl SymmetricKeyAlgorithm {
         Ok(ciphertext)
     }
 
-    pub fn encrypt_protected(self, key: &[u8], plaintext: &[u8]) -> Result<Vec<u8>> {
-        self.encrypt_protected_with_rng(&mut thread_rng(), key, plaintext)
-    }
-
     /// Encrypt the data using CFB mode, without padding. Overwrites the input.
     ///
     /// OpenPGP CFB mode uses an initialization vector (IV) of all zeros, and
@@ -497,7 +492,9 @@ impl SymmetricKeyAlgorithm {
 
         {
             match self {
-                SymmetricKeyAlgorithm::Plaintext => {}
+                SymmetricKeyAlgorithm::Plaintext => {
+                    bail!("'Plaintext' is not a legal cipher for encrypted data")
+                }
                 SymmetricKeyAlgorithm::IDEA => {
                     encrypt!(Idea, key, iv_vec, prefix, data, bs, resync)
                 }
@@ -549,7 +546,9 @@ impl SymmetricKeyAlgorithm {
     ) -> Result<()> {
         // TODO: actual cfb mode used in pgp
         match self {
-            SymmetricKeyAlgorithm::Plaintext => {}
+            SymmetricKeyAlgorithm::Plaintext => {
+                bail!("'Plaintext' is not a legal cipher for encrypted data")
+            }
             SymmetricKeyAlgorithm::IDEA => encrypt_regular!(Idea, key, iv_vec, plaintext),
             SymmetricKeyAlgorithm::TripleDES => {
                 encrypt_regular!(TdesEde3, key, iv_vec, plaintext);
@@ -579,8 +578,8 @@ impl SymmetricKeyAlgorithm {
     }
 
     /// Generate a new session key.
-    pub fn new_session_key<R: Rng + CryptoRng>(self, rng: &mut R) -> Vec<u8> {
-        let mut session_key = vec![0u8; self.key_size()];
+    pub fn new_session_key<R: Rng + CryptoRng>(self, mut rng: R) -> Zeroizing<Vec<u8>> {
+        let mut session_key = Zeroizing::new(vec![0u8; self.key_size()]);
         rng.fill_bytes(&mut session_key);
         session_key
     }
@@ -588,25 +587,23 @@ impl SymmetricKeyAlgorithm {
 
 #[cfg(test)]
 mod tests {
-    use super::*;
-
     use rand::SeedableRng;
-    use rand_xorshift::XorShiftRng;
+    use rand_chacha::ChaCha8Rng;
+
+    use super::*;
 
     macro_rules! roundtrip {
         ($name:ident, $alg:path) => {
             #[test]
             fn $name() {
-                let rng = &mut XorShiftRng::from_seed([
-                    0x3, 0x8, 0x3, 0xe, 0x3, 0x8, 0x3, 0xe, 0x3, 0x8, 0x3, 0xe, 0x3, 0x8, 0x3, 0xe,
-                ]);
+                let mut rng = ChaCha8Rng::seed_from_u64(0);
 
                 // Protected
                 for i in 1..1024 {
                     let data = (0..i).map(|_| rng.gen()).collect::<Vec<_>>();
                     let key = (0..$alg.key_size()).map(|_| rng.gen()).collect::<Vec<_>>();
 
-                    let mut ciphertext = $alg.encrypt_protected(&key, &data).unwrap();
+                    let mut ciphertext = $alg.encrypt_protected(&mut rng, &key, &data).unwrap();
                     assert_ne!(data, ciphertext);
 
                     let plaintext = $alg.decrypt_protected(&key, &mut ciphertext).unwrap();
