@@ -2,6 +2,7 @@ use std::str;
 
 use bstr::BString;
 use chrono::{DateTime, Duration, TimeZone, Utc};
+use log::{debug, warn};
 use nom::bytes::streaming::{tag, take};
 use nom::combinator::{complete, map, map_opt, map_parser, map_res, rest};
 use nom::multi::{fold_many_m_n, length_data, many0};
@@ -13,18 +14,17 @@ use crate::crypto::aead::AeadAlgorithm;
 use crate::crypto::hash::HashAlgorithm;
 use crate::crypto::public_key::PublicKeyAlgorithm;
 use crate::crypto::sym::SymmetricKeyAlgorithm;
-use crate::de::Deserialize;
 use crate::errors::{IResult, Result};
 use crate::packet::signature::types::*;
 use crate::types::{
-    mpi, CompressionAlgorithm, KeyId, KeyVersion, Mpi, MpiRef, RevocationKey, RevocationKeyClass,
-    Version,
+    mpi, CompressionAlgorithm, Fingerprint, KeyId, KeyVersion, Mpi, MpiRef, RevocationKey,
+    RevocationKeyClass, SignatureBytes, Version,
 };
 use crate::util::{clone_into_array, packet_length};
 
-impl Deserialize for Signature {
+impl Signature {
     /// Parses a `Signature` packet from the given slice.
-    fn from_slice(packet_version: Version, input: &[u8]) -> Result<Self> {
+    pub fn from_slice(packet_version: Version, input: &[u8]) -> Result<Self> {
         let (_, pk) = parse(packet_version)(input)?;
 
         Ok(pk)
@@ -41,8 +41,8 @@ fn duration_from_timestamp(ts: u32) -> Option<Duration> {
     Duration::try_seconds(i64::from(ts))
 }
 
-/// Parse a signature creation time subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.4
+/// Parse a Signature Creation Time subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-signature-creation-time
 fn signature_creation_time(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map_opt(
         // 4-octet time field
@@ -51,8 +51,8 @@ fn signature_creation_time(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Parse an issuer subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.5
+/// Parse an Issuer Key ID subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-issuer-key-id
 fn issuer(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(
         map_res(complete(take(8u8)), KeyId::from_slice),
@@ -60,8 +60,8 @@ fn issuer(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Parse a key expiration time subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.6
+/// Parse a Key Expiration Time subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-key-expiration-time
 fn key_expiration(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map_opt(
         // 4-octet time field
@@ -70,8 +70,8 @@ fn key_expiration(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Parse a preferred symmetric algorithms subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.7
+/// Parse a Preferred Symmetric Ciphers for v1 SEIPD subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#preferred-v1-seipd
 fn pref_sym_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     let list: SmallVec<[SymmetricKeyAlgorithm; 8]> = body
         .iter()
@@ -81,8 +81,27 @@ fn pref_sym_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     Ok((&b""[..], SubpacketData::PreferredSymmetricAlgorithms(list)))
 }
 
-/// Parse a preferred hash algorithms subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.8
+/// Parse a Preferred AEAD Ciphersuites subpacket (for SEIPD v2)
+///
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-preferred-aead-ciphersuites
+fn pref_aead_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
+    if body.len() % 2 != 0 {
+        return Err(nom::Err::Error(crate::errors::Error::Message(format!(
+            "Illegal preferred aead subpacket len {} must be a multiple of 2",
+            body.len(),
+        ))));
+    }
+
+    let list: SmallVec<[(SymmetricKeyAlgorithm, AeadAlgorithm); 4]> = body
+        .chunks(2)
+        .map(|v| (SymmetricKeyAlgorithm::from(v[0]), AeadAlgorithm::from(v[1])))
+        .collect();
+
+    Ok((&b""[..], SubpacketData::PreferredAeadAlgorithms(list)))
+}
+
+/// Parse a Preferred Hash Algorithms subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-preferred-hash-algorithms
 fn pref_hash_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     let list: SmallVec<[HashAlgorithm; 8]> = body
         .iter()
@@ -92,8 +111,8 @@ fn pref_hash_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     Ok((&b""[..], SubpacketData::PreferredHashAlgorithms(list)))
 }
 
-/// Parse a preferred compression algorithms subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.9
+/// Parse a Preferred Compression Algorithms subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-preferred-compression-algor
 fn pref_com_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     let list: SmallVec<[CompressionAlgorithm; 8]> = body
         .iter()
@@ -106,8 +125,8 @@ fn pref_com_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     ))
 }
 
-/// Parse a signature expiration time subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.10
+/// Parse a Signature Expiration Time subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-signature-expiration-time
 fn signature_expiration_time(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map_opt(
         // 4-octet time field
@@ -116,36 +135,36 @@ fn signature_expiration_time(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Parse a exportable certification subpacket.
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.11
+/// Parse an Exportable Certification subpacket.
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-exportable-certification
 fn exportable_certification(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(complete(be_u8), |v| {
         SubpacketData::ExportableCertification(v == 1)
     })(i)
 }
 
-/// Parse a revocable subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.12
+/// Parse a Revocable subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-revocable
 fn revocable(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(complete(be_u8), |v| SubpacketData::Revocable(v == 1))(i)
 }
 
-/// Parse a trust signature subpacket.
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.13
+/// Parse a Trust Signature subpacket.
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-trust-signature
 fn trust_signature(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(pair(be_u8, be_u8), |(depth, value)| {
         SubpacketData::TrustSignature(depth, value)
     })(i)
 }
 
-/// Parse a regular expression subpacket.
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.14
+/// Parse a Regular Expression subpacket.
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-regular-expression
 fn regular_expression(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(map(rest, BString::from), SubpacketData::RegularExpression)(i)
 }
 
-/// Parse a revocation key subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.15
+/// Parse a Revocation Key subpacket (Deprecated)
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-revocation-key-deprecated
 fn revocation_key(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(
         tuple((
@@ -160,8 +179,8 @@ fn revocation_key(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Parse a notation data subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.16
+/// Parse a Notation Data subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-notation-data
 fn notation_data(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     // Flags
     let (i, readable) = map(be_u8, |v| v == 0x80)(i)?;
@@ -181,8 +200,8 @@ fn notation_data(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     ))
 }
 
-/// Parse a key server preferences subpacket
-/// https://tools.ietf.org/html/rfc4880.html#section-5.2.3.17
+/// Parse a Key Server Preferences subpacket
+/// https://www.rfc-editor.org/rfc/rfc9580.html#name-key-server-preferences
 fn key_server_prefs(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     Ok((
         &b""[..],
@@ -190,30 +209,30 @@ fn key_server_prefs(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     ))
 }
 
-/// Parse a preferred key server subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.18
+/// Parse a Preferred Key Server subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-preferred-key-server
 fn preferred_key_server(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(map_res(rest, str::from_utf8), |body| {
         SubpacketData::PreferredKeyServer(body.to_string())
     })(i)
 }
 
-/// Parse a primary user id subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.19
+/// Parse a Primary User ID subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-primary-user-id
 fn primary_userid(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(be_u8, |a| SubpacketData::IsPrimary(a == 1))(i)
 }
 
-/// Parse a policy URI subpacket.
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.20
+/// Parse a Policy URI subpacket.
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-policy-uri
 fn policy_uri(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(map_res(rest, str::from_utf8), |body| {
         SubpacketData::PolicyURI(body.to_owned())
     })(i)
 }
 
-/// Parse a key flags subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.21
+/// Parse a Key Flags subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-key-flags
 fn key_flags(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     Ok((
         &b""[..],
@@ -221,22 +240,14 @@ fn key_flags(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     ))
 }
 
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.22
+/// Parse a Signer's User ID subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-key-flags
 fn signers_userid(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     Ok((&[], SubpacketData::SignersUserID(BString::from(i))))
 }
 
-/// Parse a features subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.24
-fn features(body: &[u8]) -> IResult<&[u8], SubpacketData> {
-    Ok((
-        &b""[..],
-        SubpacketData::Features(SmallVec::from_slice(body)),
-    ))
-}
-
-/// Parse a revocation reason subpacket
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.23
+/// Parse a Reason for Revocation subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-key-flags
 fn rev_reason(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(
         pair(map(be_u8, RevocationCode::from), map(rest, BString::from)),
@@ -244,7 +255,17 @@ fn rev_reason(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.25
+/// Parse a Features subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-features
+fn features(body: &[u8]) -> IResult<&[u8], SubpacketData> {
+    Ok((
+        &b""[..],
+        SubpacketData::Features(SmallVec::from_slice(body)),
+    ))
+}
+
+/// Parse a Signature Target subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-signature-target
 fn sig_target(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(
         tuple((
@@ -258,29 +279,59 @@ fn sig_target(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     )(i)
 }
 
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3.26
+/// Parse an Embedded Signature subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-embedded-signature
 fn embedded_sig(i: &[u8]) -> IResult<&[u8], SubpacketData> {
     map(parse(Version::New), |sig| {
         SubpacketData::EmbeddedSignature(Box::new(sig))
     })(i)
 }
 
-/// Parse an issuer subpacket
-/// Ref: https://tools.ietf.org/html/draft-ietf-openpgp-rfc4880bis-05#section-5.2.3.28
+/// Parse an Issuer Fingerprint subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-issuer-fingerprint
 fn issuer_fingerprint(i: &[u8]) -> IResult<&[u8], SubpacketData> {
-    map(
-        pair(map(be_u8, KeyVersion::from), rest),
-        |(version, fingerprint)| {
-            SubpacketData::IssuerFingerprint(version, SmallVec::from_slice(fingerprint))
-        },
-    )(i)
+    let (i, version) = map(be_u8, KeyVersion::from)(i)?;
+
+    // This subpacket is only used for v4 and newer fingerprints
+    if version != KeyVersion::V4 && version != KeyVersion::V5 && version != KeyVersion::V6 {
+        return Err(invalid_key_version(version));
+    }
+
+    if let Some(fingerprint_len) = version.fingerprint_len() {
+        let (i, fingerprint) = take(fingerprint_len)(i)?;
+        let fp = Fingerprint::new(version, fingerprint)?;
+
+        Ok((i, SubpacketData::IssuerFingerprint(fp)))
+    } else {
+        Err(invalid_key_version(version))
+    }
 }
 
-/// Parse a preferred aead subpacket
-fn pref_aead_alg(body: &[u8]) -> IResult<&[u8], SubpacketData> {
+/// Parse a preferred encryption modes subpacket (non-RFC subpacket for GnuPG "OCB" mode)
+fn preferred_encryption_modes(body: &[u8]) -> IResult<&[u8], SubpacketData> {
     let list: SmallVec<[AeadAlgorithm; 2]> = body.iter().map(|v| AeadAlgorithm::from(*v)).collect();
 
-    Ok((&b""[..], SubpacketData::PreferredAeadAlgorithms(list)))
+    Ok((&b""[..], SubpacketData::PreferredEncryptionModes(list)))
+}
+
+/// Parse an Intended Recipient Fingerprint subpacket
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-intended-recipient-fingerpr
+fn intended_recipient_fingerprint(i: &[u8]) -> IResult<&[u8], SubpacketData> {
+    let (i, version) = map(be_u8, KeyVersion::from)(i)?;
+
+    // This subpacket is only used for v4 and newer fingerprints
+    if version != KeyVersion::V4 && version != KeyVersion::V5 && version != KeyVersion::V6 {
+        return Err(invalid_key_version(version));
+    }
+
+    if let Some(fingerprint_len) = version.fingerprint_len() {
+        let (i, fingerprint) = take(fingerprint_len)(i)?;
+        let fp = Fingerprint::new(version, fingerprint)?;
+
+        Ok((i, SubpacketData::IntendedRecipientFingerprint(fp)))
+    } else {
+        Err(invalid_key_version(version))
+    }
 }
 
 fn subpacket(typ: SubpacketType, is_critical: bool, body: &[u8]) -> IResult<&[u8], Subpacket> {
@@ -312,6 +363,8 @@ fn subpacket(typ: SubpacketType, is_critical: bool, body: &[u8]) -> IResult<&[u8
         SignatureTarget => sig_target(body),
         EmbeddedSignature => embedded_sig(body),
         IssuerFingerprint => issuer_fingerprint(body),
+        PreferredEncryptionModes => preferred_encryption_modes(body),
+        IntendedRecipientFingerprint => intended_recipient_fingerprint(body),
         PreferredAead => pref_aead_alg(body),
         Experimental(n) => Ok((
             body,
@@ -339,23 +392,33 @@ fn subpackets<'a>(i: &'a [u8]) -> IResult<&'a [u8], Vec<Subpacket>> {
     }))(i)
 }
 
-fn actual_signature(typ: &PublicKeyAlgorithm) -> impl Fn(&[u8]) -> IResult<&[u8], Vec<Mpi>> + '_ {
+fn actual_signature(
+    typ: &PublicKeyAlgorithm,
+) -> impl Fn(&[u8]) -> IResult<&[u8], SignatureBytes> + '_ {
     move |i: &[u8]| match typ {
         &PublicKeyAlgorithm::RSA | &PublicKeyAlgorithm::RSASign => {
-            map(mpi, |v| vec![v.to_owned()])(i)
+            map(mpi, |v| vec![v.to_owned()].into())(i)
         }
-        &PublicKeyAlgorithm::DSA | &PublicKeyAlgorithm::ECDSA | &PublicKeyAlgorithm::EdDSA => {
-            fold_many_m_n(
-                2,
-                2,
-                mpi,
-                Vec::new,
-                |mut acc: Vec<Mpi>, item: MpiRef<'_>| {
-                    acc.push(item.to_owned());
-                    acc
-                },
-            )(i)
+        &PublicKeyAlgorithm::DSA
+        | &PublicKeyAlgorithm::ECDSA
+        | &PublicKeyAlgorithm::EdDSALegacy => fold_many_m_n(
+            2,
+            2,
+            mpi,
+            Vec::new,
+            |mut acc: Vec<Mpi>, item: MpiRef<'_>| {
+                acc.push(item.to_owned());
+                acc
+            },
+        )(i)
+        .map(|(i, sig)| (i, sig.into())),
+
+        &PublicKeyAlgorithm::Ed25519 => {
+            let (i, sig) = nom::bytes::complete::take(64u8)(i)?;
+
+            Ok((i, SignatureBytes::Native(sig.to_vec())))
         }
+
         &PublicKeyAlgorithm::Private100
         | &PublicKeyAlgorithm::Private101
         | &PublicKeyAlgorithm::Private102
@@ -366,13 +429,13 @@ fn actual_signature(typ: &PublicKeyAlgorithm) -> impl Fn(&[u8]) -> IResult<&[u8]
         | &PublicKeyAlgorithm::Private107
         | &PublicKeyAlgorithm::Private108
         | &PublicKeyAlgorithm::Private109
-        | &PublicKeyAlgorithm::Private110 => map(mpi, |v| vec![v.to_owned()])(i),
-        _ => Ok((i, vec![])), // don't assume format, could be non-MPI
+        | &PublicKeyAlgorithm::Private110 => map(mpi, |v| vec![v.to_owned()].into())(i),
+        _ => Ok((i, SignatureBytes::Native(vec![]))), // don't assume format, could be non-MPI
     }
 }
 
 /// Parse a v2 or v3 signature packet
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.2
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-version-3-signature-packet-
 fn v3_parser(
     packet_version: Version,
     version: SignatureVersion,
@@ -394,31 +457,43 @@ fn v3_parser(
             // Two-octet field holding left 16 bits of signed hash value.
             take(2usize),
         ))(i)?;
-        // One or more multiprecision integers comprising the signature.
+        // The SignatureBytes comprising the signature.
         let (i, sig) = actual_signature(&pub_alg)(i)?;
-        Ok((i, {
-            let mut s = Signature::new(
-                packet_version,
-                version,
-                typ,
-                pub_alg,
-                hash_alg,
-                clone_into_array(ls_hash),
-                sig,
-                vec![],
-                vec![],
-            );
 
-            s.config.created = Some(created);
-            s.config.issuer = Some(issuer);
-
-            s
-        }))
+        match version {
+            SignatureVersion::V2 => Ok((i, {
+                Signature::v2(
+                    packet_version,
+                    typ,
+                    pub_alg,
+                    hash_alg,
+                    created,
+                    issuer,
+                    clone_into_array(ls_hash),
+                    sig,
+                )
+            })),
+            SignatureVersion::V3 => Ok((i, {
+                Signature::v3(
+                    packet_version,
+                    typ,
+                    pub_alg,
+                    hash_alg,
+                    created,
+                    issuer,
+                    clone_into_array(ls_hash),
+                    sig,
+                )
+            })),
+            _ => Err(nom::Err::Error(crate::errors::Error::Message(
+                "must only be called for V2/V3".to_string(),
+            ))),
+        }
     }
 }
 
-/// Parse a v4 or v5 signature packet
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2.3
+/// Parse a v4 signature packet
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-versions-4-and-6-signature-
 fn v4_parser(
     packet_version: Version,
     version: SignatureVersion,
@@ -440,13 +515,20 @@ fn v4_parser(
             // Two-octet field holding the left 16 bits of the signed hash value.
             take(2usize),
         ))(i)?;
-        // One or more multiprecision integers comprising the signature.
+        // The SignatureBytes comprising the signature.
         let (i, sig) = actual_signature(&pub_alg)(i)?;
+
+        if version != SignatureVersion::V4 {
+            return Err(nom::Err::Error(crate::errors::Error::Message(format!(
+                "Unsupported version {:?}",
+                version
+            ))));
+        }
+
         Ok((
             i,
-            Signature::new(
+            Signature::v4(
                 packet_version,
-                version,
                 typ,
                 pub_alg,
                 hash_alg,
@@ -459,21 +541,82 @@ fn v4_parser(
     }
 }
 
-fn invalid_version(_body: &[u8], version: SignatureVersion) -> IResult<&[u8], Signature> {
-    Err(nom::Err::Error(crate::errors::Error::Unsupported(format!(
-        "unknown signature version {version:?}"
-    ))))
+/// Parse a v6 signature packet
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-versions-4-and-6-signature-
+fn v6_parser(packet_version: Version) -> impl Fn(&[u8]) -> IResult<&[u8], Signature> {
+    move |i: &[u8]| {
+        let (i, (typ, pub_alg, hash_alg, hsub, usub, ls_hash)) = tuple((
+            // One-octet signature type.
+            map_res(be_u8, SignatureType::try_from),
+            // One-octet public-key algorithm.
+            map(be_u8, PublicKeyAlgorithm::from),
+            // One-octet hash algorithm.
+            map(be_u8, HashAlgorithm::from),
+            // Four-octet scalar octet count for following hashed subpacket data.
+            // Hashed subpacket data set (zero or more subpackets).
+            map_parser(length_data(be_u32), subpackets),
+            // Four-octet scalar octet count for the following unhashed subpacket data.
+            // Unhashed subpacket data set (zero or more subpackets).
+            map_parser(length_data(be_u32), subpackets),
+            // Two-octet field holding the left 16 bits of the signed hash value.
+            take(2usize),
+        ))(i)?;
+
+        // A variable-length field containing:
+        // A one-octet salt size. The value MUST match the value defined for the hash algorithm as specified in Table 23.
+        // The salt; a random value of the specified size.
+        let (i, len) = be_u8(i)?;
+        let (i, salt) = take(len)(i)?;
+
+        if hash_alg.salt_len() != Some(salt.len()) {
+            return Err(nom::Err::Error(crate::errors::Error::Message(format!(
+                "Illegal salt length {} found for {:?}",
+                salt.len(),
+                hash_alg
+            ))));
+        }
+
+        // The SignatureBytes comprising the signature.
+        let (i, sig) = actual_signature(&pub_alg)(i)?;
+        Ok((
+            i,
+            Signature::v6(
+                packet_version,
+                typ,
+                pub_alg,
+                hash_alg,
+                clone_into_array(ls_hash),
+                sig,
+                hsub,
+                usub,
+                salt.to_vec(),
+            ),
+        ))
+    }
+}
+
+fn invalid_sig_version(version: SignatureVersion) -> nom::Err<crate::errors::Error> {
+    nom::Err::Error(crate::errors::Error::Unsupported(format!(
+        "invalid signature version {version:?}"
+    )))
+}
+
+fn invalid_key_version(version: KeyVersion) -> nom::Err<crate::errors::Error> {
+    nom::Err::Error(crate::errors::Error::Unsupported(format!(
+        "invalid key version {version:?}"
+    )))
 }
 
 /// Parse a signature packet (Tag 2)
-/// Ref: https://tools.ietf.org/html/rfc4880.html#section-5.2
+/// Ref: https://www.rfc-editor.org/rfc/rfc9580.html#name-signature-packet-type-id-2
 fn parse(packet_version: Version) -> impl Fn(&[u8]) -> IResult<&[u8], Signature> {
     move |i: &[u8]| {
         let (i, version) = map(be_u8, SignatureVersion::from)(i)?;
         let (i, signature) = match &version {
             &SignatureVersion::V2 | &SignatureVersion::V3 => v3_parser(packet_version, version)(i),
             &SignatureVersion::V4 | &SignatureVersion::V5 => v4_parser(packet_version, version)(i),
-            _ => invalid_version(i, version),
+            &SignatureVersion::V6 => v6_parser(packet_version)(i),
+            _ => Err(invalid_sig_version(version)),
         }?;
         Ok((i, signature))
     }
